@@ -26,16 +26,12 @@
 
 #include <glib-object.h>
 #include <gtk/gtk.h>
-#include <dbus/dbus-glib.h>
 #include <bluetooth-enums.h>
 
 G_BEGIN_DECLS
 
 GtkTreeModel *bluetooth_chooser_get_model (BluetoothChooser *self);
-GtkTreeViewColumn *bluetooth_chooser_get_device_column (BluetoothChooser *self);
-GtkTreeViewColumn *bluetooth_chooser_get_type_column (BluetoothChooser *self);
 GtkWidget *bluetooth_chooser_get_treeview (BluetoothChooser *self);
-GtkWidget *bluetooth_chooser_get_scrolled_window (BluetoothChooser *self);
 gboolean bluetooth_chooser_remove_selected_device (BluetoothChooser *self);
 
 G_END_DECLS
